% Code for Example 4.2 in the manuscript.
%
% Computes sqrt(A)*b using rational approximations
% for sqrt provided in Hale, Higham and Trefethen's paper
% "Computing A^?,\log(A), and related matrix functions by 
% contour integrals", SINUM 2008 (Methods 2 and 3),
% combined with structured system solving.
%
% The test matrix A is taken from Simoncini, "Extended Krylov 
% subspace for parameter dependent systems", Appl. Num. Math. 2010,
% Example 3.3. This is the centered finite difference discretization
% of the operator -(u_xx + u_yy) + 10 u_x.
% It is real symmetric and positive definite, of size 2500x2500.
% 
% A can be seen as
% - either a (50,50)-quasiseparable matrix with scalar blocks,
% - or a (1,1)-quasiseparable matrix with 50x50 blocks.
%

% Using the Schwarz-Christoffel toolbox by T.A. Driscoll.
addpath ./3Nicks/sc/  % path for SC toolbox (modify as needed)
 
% Form A.

n=50;
N=2500; % matrix size

disp('Form matrices...')

% form 2D negative Laplacian L
D=toeplitz([4;-1;zeros(n-2,1)],[4,-1,zeros(1,n-2)]);
L=zeros(n^2);
for j=1:n
    L(n*(j-1)+1:n*j,n*(j-1)+1:n*j)=D;
end
for j=1:n-1
    L(n*(j-1)+1:n*j,n*j+1:n*(j+1))=-eye(n);
    L(n*j+1:n*(j+1),n*(j-1)+1:n*j)=-eye(n);
end

% now write matrix B for du/dx
step=1/(n+1); 
E=diag(ones(n-1,1),1)-diag(ones(n-1,1),-1);
B=zeros(n^2);
for j=1:n
    B(n*(j-1)+1:j*n,n*(j-1)+1:j*n)=E;
end
A=L+5*step*B; 

% Compute generators for A (1x1 blocks, quasiseparable order 50).
n2=n^2;
d=4*ones(n2,1);  % main diagonal
% form lower generators
p=[-ones(n2,1),zeros(n2,n-1)];
q=[(1+5/(n+1))*ones(1,n2-1);zeros(n-2,n2-1);ones(1,n2-1)];
q(1,n:n:n2-1)=0;
a=zeros(n,n,n2-1);
for j=2:n2-1
    a(:,:,j)=[zeros(n,1),[eye(n-1);zeros(1,n-1)]];
end
% form upper generators
g=[(1-5/(n+1))*ones(n2-1,1),zeros(n2-1,n-2),ones(n2-1,1)];
g(n:n:n2-1,1)=0;
h=[-ones(1,n2);zeros(n-1,n2)];
b=zeros(n,n,n2-1);
for k=2:n2-1
    b(:,:,k)=[zeros(n,1),[eye(n-1);zeros(1,n-1)]]';
end

% Compute other generators for A (50x50 blocks, quasiseparable order 1).
d_block=repmat(D+5*step*E,1,1,n);
p_block=-repmat(eye(n),1,1,n);
q_block=repmat(eye(n),1,1,n-1);
a_block=zeros(n,n,n);
g_block=-repmat(eye(n),1,1,n-1);
h_block=repmat(eye(n),1,1,n);
b_block=zeros(n,n,n);

disp('...done.')
%%%%%%%%%%%%%%%%%%%%%%%%%%%

I = eye(size(A));
e = eig(A);
me = min(e); Me = max(e); % only for toy problems; should be replaced with estimation?

gvec = ones(N,1); % (this would be b)
gvec=gvec/norm(gvec);

disp('Computing "exact" solution for comparison...')
c = sqrtm(A)*gvec; % for comparison
disp('...done.')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Apply Method 2 to A with (50,50)-q.s. structure
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

k = ((Me/me)^(1/4)-1)/((Me/me)^(1/4)+1);
L = -log(k)/pi;
[K,Kp] = ellipkkp(L);

% print errors
fprintf('\n\n%s\n','Method 2 with structured system solving, (50,50)-q.s. structure')
fprintf('\n%s %s %s\n','n_iter','err_class','err_struct')

for M=6:20 % number of integration nodes 

t = .5i*Kp - K + (.5:M)*2*K/M;
[u cn dn] = ellipjc(t,L);
w = (me*Me)^(1/4)*((1/k+u)./(1/k-u));
dzdt = cn.*dn./(1/k-u).^2;
% compute solution vector v using classical approach
v = zeros(N,1);
for j = 1:M
    v = v + (sqrt(w(j)^2)/w(j))*((w(j)^2*I-A)\gvec)*dzdt(j);
end
v = -8*K*(me*Me)^(1/4)*A*imag(v)/(k*pi*M);
error = norm(c-v)/norm(c);
% compute solution vector v_struct using structured approach
v_struct = zeros(N,1);
sigma=-w.^2;
x=qs_shifted_function(p,q,a,g,h,b,d,sigma,gvec);
for j = 1:M
    v_struct = v_struct - (sqrt(w(j)^2)/w(j))*x(:,j)*dzdt(j);
end
v_struct = -8*K*(me*Me)^(1/4)*A*imag(v_struct)/(k*pi*M);
error_struct = norm(c-v_struct)/norm(c);

fprintf('%4d %10.2e %10.2e\n', M, error, error_struct)

end




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Apply Method 3 to A with (50,50)-q.s. structure
% and with block (1,1)-q.s. structure
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

k2 = me/Me; % elliptic functions parameter k^2

gvec_block=ones(50,50);
gvec_block=gvec_block/norm(gvec_block,'fro');

L=-.5*log(k2)/pi;
[K,Kp]=ellipkkp(L);

% print errors
fprintf('\n\n%s\n','Method 3 with structured system solving')
fprintf('\n%s %s %s %s\n','n_iter','err_class','err_struct','err_struct_block')

for M=6:15

t = 1i*(.5:M)*Kp/M;

[sn cn dn] = ellipjc(t,L);

w = sqrt(me)*sn;
dzdt = cn.*dn;
% compute solution vector v using classical approach
% (this could also be made faster using sparse solver)

v = zeros(N,1);
for j = 1:M
    v = v - (A-w(j)^2*I)\gvec*dzdt(j);
end
v = (-2*Kp*sqrt(me)/(pi*M))*A*v;

error = norm(c-v)/norm(c);

% compute solution vector v_struct using structured approach

v_struct = zeros(N,1);
sigma=-w.^2;
x=qs_shifted_function(p,q,a,g,h,b,d,sigma,gvec);
for j = 1:M
    v_struct = v_struct - x(:,j)*dzdt(j);
end
v_struct = (-2*Kp*sqrt(me)/(pi*M))*A*v_struct;

error_struct = norm(c-v_struct)/norm(c);

% compute solution vector v_struct_block using block structured approach

v_struct_block = zeros(N,1);
x=qs_shifted_function_blocks(p_block,q_block,a_block,g_block,h_block,...
    b_block,d_block,sigma,gvec_block);
for j = 1:M
    v_struct_block = v_struct_block - x(:,j)*dzdt(j);
end
v_struct_block = (-2*Kp*sqrt(me)/(pi*M))*A*v_struct_block;
error_struct_block = norm(c-v_struct_block)/norm(c);


fprintf('%4d %10.2e %10.2e %10.2e\n', M, error, error_struct, error_struct_block)
end

