function x=qs_shifted_function_Sylvester(p,q,a,g,h,b,d,sigma,y)

% This is companion software to the manuscript
% "Efficient solution of parameter dependent
% quasiseparable systems",
% by P. Boito, Y. Eidelman and L. Gemignani.
%
% This code computes the solution of shifted linear systems
% of the form (A+sigma_i I)x_i = y, i=1,...,M,
% where A is quasiseparable.
% This version is adapted to work with a different
% rhs y for each system.
%
% We assume a scalar qs structure for now
% (i.e., m_i=1 for all indices i), although
% the m_i's still come up every now and then.

% Input:
% - lower generators p,q,a
% - upper generators g,h,b
% - diagonal entries d
% - shifts sigma.
% - rhs matrix y.
% Output:
% - solution vectors x.
%
% Please note: no check is made on the consistency of the sizes
% of the input arrays.
%
% Last modification on Oct 20, 2016.

% Set up initial data.


M=length(sigma);  % number of shifts
[N,rL]=size(p); % N = size of system, rL = lower qs order
[rU,~]=size(h); % upper qs order
% y should be NxM (its columns are the rhs vectors)


% p is N x rL (but first row not used)
% q is rL x (N-1)
% a is rL x rL x (N-1)  but first matrix not used
% g is (N-1) x rU
% h is rU x N but first column not used
% b is rU x rU x (N-1)  but first matrix not used
% d is Nx1
% y is Nx1
% sigma is Mx1


% Initialize matrix of solutions.
x=zeros(N,M);
w=zeros(N,1);   % intermediate solution

% Form full matrix (for checking computations).
% A=recover_qs(p,q,a,g,h,b,d);
% As=A+sigma(1)*eye(N);
% AA=A; % for later purposes
% yy=y;

% Implement algorithm.

% Start with the factorization A + sigma I = V * T:
% initialize generators for V
pV=zeros(N,rL);
qV=zeros(rL,N-1);
aV=zeros(rL,rL,N-1);
%gV=zeros(N-1,rU);
%hV=zeros(rU,N);
%bV=zeros(rU,rU,N-1);
dV=zeros(N,1);
c=zeros(N,M);

% 1 %

% Compute sizes:
m=ones(N,1); % block sizes
rho=zeros(N,1);
rhop=zeros(N,1);
nu=zeros(N,1);
for k=N:-1:2
    rho(k-1)=min(m(k)+rho(k),rL);
end
for k=1:N-1
    rhop(k)=rho(k)+rU;
end
nu(1)=m(1)+rho(1);
for k=2:N
    nu(k)=m(k)+rho(k)-rho(k-1);
end

gT=zeros(N-1,rhop(1));
hT=zeros(rhop(1),N);
bT=zeros(rhop(1),rhop(1),N-1);
dT=zeros(N,1);

V_N=1;
Xold=p(N,:);
% dV(N) is empty
pV(N,1)=1; % because pV_N=V_N;
%hT(1:rU,N)=h(:,N);
%dT(N)=d(N); % because V_N=1
hT(1:rhop(N-1),N)=[h(:,N);V_N'*d(N)];


c=y(N,:);
% w(N) is empty
gamma_N=[zeros(rU,1);pV(N,1)'];
gamma=zeros(2*rU,N-1);

for k=N-1:-1:2
    % 1.2.1
    emme=[p(k,:);Xold*a(:,:,k)];
    [V,R]=qr(emme);
    % check %%%%
    %      A(k:k+rho(k),:)=V'*A(k:k+rho(k),:);
    %      As(k:k+rho(k),:)=V'*As(k:k+rho(k),:);
    %      yy(k:k+rho(k))=V'*yy(k:k+rho(k));
    %%%%%%%%%%%%
    Xnew=R(1:rho(k-1),:);
    pV(k,1:rho(k-1))=V(1,1:rho(k-1));
    aV(1:rho(k),1:rho(k-1),k)=V(2:2+rho(k)-1,1:rho(k-1));
    if rho(k)+1>rho(k-1)
        dV(k)=V(1,rho(k-1)+1:rho(k-1)+nu(k));
        qV(:,k)=V(2:2+rho(k)-1,rho(k-1)+1);
    end
    
    %1.2.2
    W=[h(:,k),b(:,:,k),zeros(rU,rho(k));d(k),g(k,:),zeros(1,rho(k));...
        Xold*q(:,k),zeros(rho(k),rU),eye(rho(k))];
    Z=[eye(rU),zeros(rU,rho(k)+1);zeros(rho(k)+1,rU),V']*W; % size (rU+rho(k)+1)x(rU+rho(k)+1)
    hT(1:rhop(k-1),k)=Z(1:rhop(k-1),1);
    bT(1:rhop(k-1),1:rhop(k),k)=Z(1:rhop(k-1),2:rhop(k)+1);
    if (rU+rho(k)+1)>rhop(k-1)
        dT(k)=Z(rhop(k-1)+1,1);
        gT(k,:)=Z(rhop(k-1)+1,2:rhop(k)+1);
    end
    
    gamma(:,k)=[zeros(rU,1);pV(k,:)'];
    for i=1:M
        auxv=V'*[y(k,i);c(i)];
        c(i)=auxv(1:rho(k-1));
        w(k,i)=auxv(end);
    end
    Xold=Xnew;
end
% 1.3 (larger sizes here)
V1=eye(nu(1));
dT1=[d(1);Xnew*q(:,1)];
gT1=[g(1,:) zeros(1,rho(1)); zeros(rho(1),rU) eye(rho(1))];
gamma1=[1;zeros(rho(1),1)];
for i=1:M
    w1(:,i)=V1'*[y(1,i);c(:,i)];  %3x1
end

% 2 %

for i=1:M   % for each shifted system
    
    dR=zeros(N,1);
    gR=zeros(N-1,max(rhop));
    v=zeros(N,1);
    
    % check %%%%%%%%%
    % B=A+sigma(i)*eye(N);
    %%%%%%%%%%%%%%%%%%
    
    % 2.1.1
    [U,R]=qr(dT1+sigma(i)*gamma1);
    dR(1)=R(1,1);
    % check %%%%%%%%%
    %    nlines=length(U);
    %    B(1:nlines,:)=U'*B(1:nlines,:);
    %    As(1:nlines,:)=U'*As(1:nlines,:);
    %%%%%%%%%%%%%%%%%%
    auxv=U'*gT1;
    auxw=U'*w1(:,i);
    gR(1,1:rhop(1))=auxv(1,:);
    Y=auxv(2:end,:);
    v(1)=auxw(1);
    alpha=auxw(2:end);
    
    % 2.1.2
    
    for k=2:N-1
        if (rU+rho(k)+1)>rhop(k-1)  % that is, if dT(k) and dV(k) are not empty
            emme=[Y*(hT(:,k)+sigma(i)*gamma(:,k));dT(k)+sigma(i)*dV(k).'];
        else
            emme=Y*(hT(1:rhop(k-1),k)+sigma(i)*gamma(1:rhop(k-1),k));
        end
        [U,R]=qr(emme);
        dR(k)=R(1,1);
        % check %%%%%%%%%
        %        nlines=length(U); % should be rho(k)+1
        %        B(k:k+nlines-1,:)=U'*B(k:k+nlines-1,:);
        %        As(k:k+nlines-1,:)=U'*As(k:k+nlines-1,:);
        %%%%%%%%%%%%%%%%%%
        if (rU+rho(k)+1)>rhop(k-1) % that is, if gT is not empty
            auxv=U'*[Y*bT(:,:,k); gT(k,:)];
            auxw=U'*[alpha; w(k,i)];
        else
            auxv=U'*Y*bT(1:rhop(k-1),1:rhop(k),k);
            auxw=U'*alpha;
        end
        
        gR(k,1:rhop(k))=auxv(1,1:rhop(k));
        v(k)=auxw(1);
        Y=auxv(2:1+rho(k),1:rhop(k));
        alpha=auxw(2:end);
    end
    
    %2.1.3
    [U,R]=qr(Y*(hT(1:rhop(N-1),N)+sigma(i)*gamma_N));
    dR(N)=R(1);
    v(N)=U'*alpha;
    
    %2.2 (compute solution x_i=x(:,i))
    
    %2.2.1
    eta=zeros(rhop(1),N-1);
    x(N,i)=dR(N)\v(N);
    eta(1:rhop(N-1),N-1)=(hT(1:rhop(N-1),N)+sigma(i)*gamma_N)*x(N,i);
    %2.2.2
    
    for k=N-1:-1:2
        x(k,i)=dR(k)\(v(k)-gR(k,1:rhop(k))*eta(1:rhop(k),k));
        eta(1:rhop(k-1),k-1)=bT(1:rhop(k-1),1:rhop(k),k)*eta(1:rhop(k),k)+...
            (hT(1:rhop(k-1),k)+sigma(i)*gamma(1:rhop(k-1),k))*x(k,i);
    end
    %2.2.3
    x(1,i)=dR(1)\(v(1)-gR(1,:)*eta(:,1));
end


end